﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Reflection;
using System.Runtime.Serialization;
using System.Text;
using System.Threading.Tasks;
using SDON.Serialization;

namespace SDON
{
    /// <summary>
    /// The root object of a SDON document page.
    /// </summary>
    [Serializable]
    public class DocumentBase : SDONSerializeable
    {
        /// <summary>
        /// Type of diagram SDON will create. Sets the behavior of lines and shapes. A value from the SDONTemplates enum
        /// </summary>
        [DataMember(Name = "Template")]
        [IgnoreIfDefaultValue(Default = null)]
        protected string _template = null;

        /// <summary>
        /// The version of the SDON document
        /// </summary>
        [DataMember(Name = "Version")]
        protected string _version = "20";

        /// <summary>
        /// The root shape in a diagram.
        /// </summary>
        [DataMember(Name = "Shape")]
        [IgnoreIfDefaultValue(Default = null)]
        protected Shape _shape = null;

        /// <summary>
        /// A title string centered over the diagram 1/2" above it
        /// </summary>
        [DataMember(Name = "Title")]
        [IgnoreIfDefaultValue(Default = null)]
        protected Title _title = null;

        /// <summary>
        /// A list of segmented lines that link shapes together
        /// </summary>
        [DataMember(Name = "Returns")]
        [IgnoreIfDefaultValue(Default = null)]
        protected List<Return> _returns = null;

        /// <summary>
        /// A list of mappings of color aliases to color hex codes so that the alias can be used instead of the hex color code. 
        /// </summary>
        [DataMember(Name = "Colors")]
        [IgnoreIfDefaultValue(Default = null)]
        protected List<ColorEntry> _colors = null;

        /// <summary>
        /// A list of mapping of a symbol's ID (in the SmartDraw CMS) to an alias that can be used in place of the Shape's ShapeType name.
        /// </summary>
        [DataMember(Name = "Symbols")]
        [IgnoreIfDefaultValue(Default = null)]
        protected List<SymbolEntry> _symbols = null;

        /// <summary>
        /// Definitions of data tables used in the diagram whose rows can be referenced by a shape. 
        /// </summary>
        [DataMember(Name = "DataTable")]
        [IgnoreIfDefaultValue(Default = null)]
        protected List<DataTable> _dataTable = null;

        /// <summary>
        /// Special-purpose data table
        /// </summary>
        [DataMember(Name = "UseDataTable")]
        [IgnoreIfDefaultValue(Default = null)]
        protected DataTableInstance _useDataTable = null;

        /// <summary>
        /// Document-wide defaults
        /// </summary>
        [DataMember(Name = "DocumentDefaults")]
        [IgnoreIfDefaultValue(Default = null)]
        protected DocumentDefaults _documentDefaults = null;

        /// <summary>
        /// Set the template  
        /// </summary>
        /// <param name="template">New template name to set</param>
        /// <returns>dynamic... Will return superclass</returns>
        public dynamic SetTemplate(string template)
        {
            _template = template;

            return (Document)this;
        }

        /// <summary>
        /// Return the main/root shape of a document.   
        /// </summary>
        /// <returns>Shape</returns>
        public Shape GetTheShape()
        {
            return _shape;
        }

        /// <summary>
        /// Set the title 
        /// </summary>
        /// <param name="title">Text of title
        /// <returns>Title</returns>
        public Title AddTitle(string title)
        {
            if ((title == null) || (title.Length == 0))
                return null;

            if (_title == null)
                _title = new Title();

            _title.SetLabel(title);

            return _title;
        }

        /// <summary>
        /// Add a return. Returns list of segmented lines that link shapes together.
        /// </summary>
        /// <param name="startID">ID of shape to start return from</param>
        /// <param name="endID">ID of shape at which to end return</param>
        /// <returns>Return</returns>
        public Return AddReturn(int startID, int endID)
        {
            if (_returns == null)
                _returns = new List<Return>();

            Return ret = new Return();

            ret.StartID = startID;
            ret.EndID = endID;

            _returns.Add(ret);

            return ret;
        }

        /// <summary>
        /// Set the color mappings of color aliases to color RBG codes.  
        /// </summary>
        /// <param name="colors">Color mappings</param>
        /// <returns>dynamic... Will return superclass</returns>
        public dynamic SetColors(List<ColorEntry> colors)
        {
            if (_colors == null)
                _colors = new List<ColorEntry>();

            _colors.AddRange(colors);

            return (Document)this;
        }

        /// <summary>
        /// Set the symbol id mappings. SDON API V2
        /// </summary>
        /// <param name="symbols">Symbol list</param>
        /// <returns>dynamic... Will return superclass</returns>
        public dynamic SetSymbols(List<SymbolEntry> symbols)
        {
            if (_symbols == null)
                _symbols = new List<SymbolEntry>();

            _symbols.AddRange(symbols);

            return (Document)this;
        }


        /// <summary>
        /// Add a data table
        /// </summary>
        /// <param name="tableName">Name of table</param>
        /// <param name="id">Id of table</param>
        /// <returns>DataTable</returns>
        public DataTable AddDataTable(string tableName, string id)
        {
            if (_dataTable == null)
                _dataTable = new List<DataTable>();

            DataTable dataTable = new DataTable();

            dataTable.ID = id;
            dataTable.TableName = tableName;

            _dataTable.Add(dataTable);

            return dataTable;
        }

        /// <summary>
        /// Get a data table by its ID    
        /// </summary>
        /// <param name="id">String form of table ID</param>
        /// <returns>DataTable</returns>
        public DataTable GetDataTable(string id)
        {
            return _dataTable.Find(table => table.ID == id);
        }


        /// <summary>
        /// Set vertical spacing for connectors
        /// </summary>
        /// <param name="verticalSpacing"></param>spacing
        /// <returns>dynamic... Will return superclass</returns>
        public dynamic SetConnectorVerticalSpacing(double verticalSpacing)
        {
            if (_documentDefaults == null)
                _documentDefaults = new DocumentDefaults();

            _documentDefaults.VerticalSpacing = verticalSpacing;

            return this;
        }

        /// <summary>
        /// Set horizontal spacing for connectors
        /// </summary>
        /// <param name="horizontalSpacing"></param>spacing
        /// <returns>dynamic... Will return superclass</returns>
        public dynamic SetConnectorHorizontalSpacing(double horizontalSpacing)
        {
            if (_documentDefaults == null)
                _documentDefaults = new DocumentDefaults();

            _documentDefaults.HorizontalSpacing = horizontalSpacing;

            return this;
        }

        public string ToJSON()
        {
            return SDON.SDONBuilder.ToJSON(this);
        }

        /// <summary>
        /// Default constructor.
        /// </summary>
        public DocumentBase()
        {
            _version = "20";
            _shape = new Shape();
        }

        /// <summary>
        /// Special constructor for implementing ISerializeable. Deserializes the object.
        /// </summary>
        /// <param name="info">Serialization info (all the values that are on the incoming object graph) of the object being deserialized.</param>
        /// <param name="context">The deserialization context.</param>
        internal DocumentBase(SerializationInfo info, StreamingContext context)
            : base(info, context)
        {
        }

        protected override void Serialize(List<MemberInfo> members, SerializationInfo info, ref StreamingContext context, ref List<string> usedPropertyNames)
        {
            base.Serialize(members, info, ref context, ref usedPropertyNames);
        }
    }
}
